{% extends "django_ca/admin/extensions/base/base.html" %}
{% load i18n django_ca %}

{% block content %}
{% for entry in extension %}
  Distribution Point:
  <ul>
    {% if entry.full_name %}
    <li>Full Name: {{ entry.full_name|format_general_names|join:", " }}</li>
    {% endif %}
    {% if entry.relative_name %}
    <li>Relative Name: {{ entry.relative_name|format_relative_name }}</li>
    {% endif %}
    {% if entry.crl_issuer %}
    <li>CRL Issuer: {{ entry.crl_issuer|format_general_names|join:", " }}</li>
    {% endif %}
    {% if entry.reasons %}
    <li>Reasons: {{ entry.reasons|join:", " }}</li>
    {% endif %}
  </ul>
{% endfor %}
{% endblock content %}
