{% load i18n %}
<span class="django-ca-extension">
  {% if extension.critical %}
  <div class="django-ca-extension-critical">
    <img src="/static/admin/img/icon-yes.svg" alt="{% trans 'Critical' %}"> {% trans 'Critical' %}
  </div>
  {% endif %}
  <div class="django-ca-extension-value">
    {% if extension is not None %}
      {% block content %}{{ extension.as_text }}{% endblock content %}
    {% else %}
      {# Extension could be None for SubjectAltName, which is always displayed even if not present #}
      {% trans "&lt;Not present&gt;" %}
    {% endif %}
  </div>
</span>
