# Generated by Django 1.10.5 on 2017-03-07 18:39

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_ca', '0004_auto_20170304_1442'),
    ]

    operations = [
        migrations.AlterField(
            model_name='certificate',
            name='cn',
            field=models.CharField(max_length=128, verbose_name='CommonName'),
        ),
        migrations.AlterField(
            model_name='certificate',
            name='serial',
            field=models.CharField(max_length=64, unique=True),
        ),
        migrations.AlterField(
            model_name='certificateauthority',
            name='cn',
            field=models.CharField(max_length=128, verbose_name='CommonName'),
        ),
        migrations.AlterField(
            model_name='certificateauthority',
            name='serial',
            field=models.CharField(max_length=64, unique=True),
        ),
    ]
