{% extends "django_ca/admin/extensions/base/base.html" %}
{% load i18n django_ca %}

{% block content %}
<ul>
{% for value in extension %}<li>{{ value }}</li>{% endfor %}
</ul>
{% endblock content %}
