# Generated by Django 1.11 on 2017-05-05 12:51

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_ca', '0005_auto_20170307_1839'),
    ]

    operations = [
        migrations.AlterField(
            model_name='certificate',
            name='csr',
            field=models.TextField(blank=True, verbose_name='CSR'),
        ),
    ]
