{% extends "django_object_actions/change_form.html" %}
{% load django_ca i18n admin_urls %}

{% block breadcrumbs %}
<div class="breadcrumbs">
<a href="{% url 'admin:index' %}">{% trans 'Home' %}</a>
&rsaquo; <a href="{% url 'admin:app_list' app_label=opts.app_label %}">{{ opts.app_config.verbose_name }}</a>
&rsaquo; {% if has_view_permission %}<a href="{% url opts|admin_urlname:'changelist' %}">{{ opts.verbose_name_plural|capfirst }}</a>{% else %}{{ opts.verbose_name_plural|capfirst }}{% endif %}
{% if object_action %}
  &rsaquo;
	{% if original_obj %}
		{% if perms.django_ca.change_certificate %}<a href="{% url opts|admin_urlname:'change' original_obj.pk %}">{{ original_obj|truncatewords:"18" }}</a>{% else %}{{ original_obj|truncatewords:"18" }}{% endif %}
	{% else %}
		{% if perms.django_ca.change_certificate %}<a href="{% url opts|admin_urlname:'change' object_id %}">{{ object|truncatewords:"18" }}</a>{% else %}{{ object|truncatewords:"18" }}{% endif %}
	{% endif %}
	&rsaquo; {{ object_action }}
{% else %}
&rsaquo; {% if add %}{% blocktrans with name=opts.verbose_name %}Add {{ name }}{% endblocktrans %}{% else %}{{ original|truncatewords:"18" }}{% endif %}
{% endif %}
</div>
{% endblock %}

{% block submit_buttons_bottom %}{% django_ca_certificate_submit_row %}{% endblock %}
