# Generated by Django 2.2 on 2019-04-27 16:26

from django.db import migrations, models
import django_ca.models


class Migration(migrations.Migration):

    dependencies = [
        ('django_ca', '0012_auto_20190405_2345'),
    ]

    operations = [
        migrations.AddField(
            model_name='certificateauthority',
            name='crl_number',
            field=models.TextField(blank=True, default='{"scope": {}}', help_text='Data structure to store the CRL number (see RFC 5280, 5.2.3) depending on the scope.', validators=[django_ca.models.json_validator], verbose_name='CRL Number'),
        ),
    ]
